#!/usr/bin/env python3
#
# Copyright 2019 PSB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""analysis_fba_comparison.py

Command-line interface for the running of two FVAs, one for
the unsMOMENTed original SBML model, and one for the sMOMENTed
one.
It uses cobrapy's FVA functions.
"""

# IMPORTS
# External module for command-line interfaces
import click

# Internal module, contains the actual FBA comparison function
from .submodules.fva_comparison import fva_comparison_with_sbml


# Set-up command-line parameters using click decorators
@click.command()
@click.option(
    "--sbml_original_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True),
    prompt="Original SBML path",
    help="Full SBML path of original model without protein allocation constraints",
)
@click.option(
    "--sbml_protein_constrained_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True),
    prompt="SBML path of sMOMENT-enhanced model",
    help="Full SBML path of sMOMENT-enhanced model.",
)
@click.option(
    "--objective",
    required=True,
    type=str,
    prompt="Objective",
    help="Objective of the comparative FVA.",
)
# Command-line interface function
def fba_comparison_cli(
    sbml_original_path: str, sbml_protein_constrained_path: str, objective: str
) -> None:
    """FVA (Flux Variability Analysis) comparison with the given arguments, e.g. in order to check the validity of the sMOMENTed model.

    An FVA result summary is shown for the original SBML model as
    well as for the protein-allocation-constrained one. The output is generated by cobrapy.

    Example: Print a comparative FVA for the non-constrained model 'C:\\original.xml' and the sMOMENT-model 'C:\\pac.xml' and the objective 'ACALD':
    python analysis_fva_comparison.py --sbml_original_path C:\\original.xml --sbml_protein_constrained_path C:\\pac.xml --objective ACALD
    """
    fva_comparison_with_sbml(
        sbml_original_path, sbml_protein_constrained_path, objective
    )


# Start-up routine if script is called
if __name__ == "__main__":
    # Thanks to the click decorators, the command-line interface
    # function does not need to be called directly. The given
    # console arguments are added automatically.
    fba_comparison_cli()
    pass
