#!/usr/bin/env python3
#
# Copyright 2019 PSB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""data_parse_sabio_rk_for_model.py

CLI for the retrieval of kcats from the SABIO-RK database
(http://sabiork.h-its.org/) and storage of them into a
machine-readable JSON.
"""

# IMPORTS
# External modules
import click

# Internal modules
from .submodules.parse_sabio_rk_for_model import parse_sabio_rk_for_model_with_sbml


# Set-up command-line parameters using click decorators
@click.command()
@click.option(
    "--sbml_path",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True),
    prompt="SBML path",
    help="The path to the metabolic model SBML from which the EC numbers will be "
    "read out in order to search them in SABIO-RK.",
)
@click.option(
    "--json_output_path",
    required=True,
    type=click.Path(dir_okay=True),
    prompt="JSON output path",
    help="The path to the newly generated JSON that will include the kcat data from SABIO-RK.",
)
@click.option(
    "--bigg_id_name_mapping_path",
    required=True,
    type=click.Path(dir_okay=True),
    prompt="BIGG ID<->Name mapping JSON path",
    help="The path to the JSON ",
)
# Command-line interface function
def parse_sabio_rk_for_model_with_sbml_cli(
    sbml_path: str, json_output_path: str, bigg_id_name_mapping_path: str
) -> None:
    """Using the given model, its EC numbers are used for a search in the SABIO-RK database. The kcat results are stored as a JSON :D

    SABIO-RK is called using its API, so that a woking internet connection is needed in order to run this script.
    In addition to the metbolic model, a BIGG ID name JSON generated by data_parse_bigg_metabolites_file.py has to be given as input.
    The resulting JSON includes found kcats for the EC numbers, including EC numbers for which kcats were only found with
    wilcard searches and EC nubmers for which a direct result was found.
    Additionally, the resulting JSON includes the organism of the EC-number-associated kcats and the substrates of the kcat
    data, using BIGG ID names if possible.

    Example
    ----------
    Create a SABIO-RK JSON called 'C:\\folder\\sabio_rk.json' using the model 'C:\\folder\\model.xml' and the
    BIGG ID<->name mapping JSON 'C:\\folder\\bigg_ids.json'
    <pre>
    python data_parse_sabio_rk_for_model.py --sbml_path C:\\folder\\model.xml --json_output_path C:\\folder\\sabio_rk.json --bigg_id_name_mapping_path C:\\folder\\bigg_ids.json
    </pre>
    """
    parse_sabio_rk_for_model_with_sbml(
        sbml_path, json_output_path, bigg_id_name_mapping_path
    )


# Start-up routine if script is called
if __name__ == "__main__":
    # Thanks to the click decorators, the command-line interface
    # function does not need to be called directly. The given
    # console arguments are added automatically.
    parse_sabio_rk_for_model_with_sbml_cli()
    pass
